# File viewer tab for bfit
# Derek Fujimoto
# Nov 2017

from tkinter import *
from tkinter import ttk
import numpy as np
import sys,os,datetime,time
from bdata import bdata
import matplotlib.pyplot as plt
from multiprocessing import Process, Pipe
from bfit.gui.zahersCalculator import current2field

__doc__ = """
    View file contents tab.
    
    To-do:
        2e mode viewing
        NBM viewing
        cumulative count viewing
    """

# =========================================================================== #
class fileviewer(object):
    """
        Data fields:
            year: year of exp
            runn: run number
            entry_asym_type: combobox for asym calculations
            text: Text widget for displaying run information
            bfit: bfit object
            fig_list: list of figures
            asym_type: drawing style
            is_updating: True if update draw
            data: bdata object for drawing
    """
    
    asym_dict = {"Combined Helicity"        :'c',
                 "Split Helicity"           :'h',
                 "Matched Helicity"         :'hm',
                 "Shifted Helicity"         :'hs',
                 "Matched Peak Finding"     :'hp',
                 "Raw Scans"                :'r',
                 "Combined Hel Raw"         :'2e_rw_c',
                 "Combined Hel Slopes"      :'2e_sl_c',
                 "Combined Hel Diff"        :'2e_di_c',
                 "Split Hel Raw"            :'2e_rw_h',
                 "Split Hel Slopes"         :'2e_sl_h',
                 "Split Hel Diff"           :'2e_di_h',
                 "Alpha Diffusion"          :'ad',
                 "Combined Hel (Alpha Tag)" :"at_c",
                 "Split Hel (Alpha Tag)"    :"at_h",
                 "Combined Hel (!Alpha Tag)":"nat_c",
                 "Split Hel (!Alpha Tag)"   :"nat_h",
                 }
    default_export_filename = "%d_%d.csv" # year_run.csv
    
    # ======================================================================= #
    def __init__(self,file_tab,bfit):
        """ Position tab tkinter elements"""
        
        # year and filenumber entry ------------------------------------------
        entry_frame = ttk.Frame(file_tab,borderwidth=1)
        self.year = StringVar()
        self.runn = StringVar()
        self.rebin = IntVar()
        self.bfit = bfit
        
        self.year.set(datetime.datetime.now().year)
        self.rebin.set(1)
        
        entry_year = ttk.Entry(entry_frame,\
                textvariable=self.year,width=5)
        self.entry_runn = ttk.Entry(entry_frame,\
                textvariable=self.runn,width=7)
        
        # fetch button
        fetch = ttk.Button(entry_frame,text='Fetch',command=self.get_data)
            
        # draw button
        draw = ttk.Button(entry_frame,text='Draw',command=self.draw)
        
        # grid and labels
        entry_frame.grid(column=0,row=0,sticky=N)
        ttk.Label(entry_frame,text="Year:").grid(column=0,row=0,sticky=E)
        entry_year.grid(column=1,row=0,sticky=E)
        ttk.Label(entry_frame,text="Run Number:").grid(column=2,row=0,sticky=E)
        self.entry_runn.grid(column=3,row=0,sticky=E)
        fetch.grid(column=4,row=0,sticky=E)
        draw.grid(column=5,row=0,sticky=E)
        
        # padding 
        for child in entry_frame.winfo_children(): 
            child.grid_configure(padx=5, pady=5)

        # viewer frame -------------------------------------------------------
        view_frame = ttk.Frame(file_tab,borderwidth=2)
        
        self.text_nw = Text(view_frame,width=75,height=20,state='normal')
        self.text_ne = Text(view_frame,width=75,height=20,state='normal')
        self.text_sw = Text(view_frame,width=75,height=20,state='normal')
        self.text_se = Text(view_frame,width=75,height=20,state='normal')
        
        ttk.Label(view_frame,text="Run Info").grid(column=0,row=0,sticky=N)
        ttk.Label(view_frame,text="PPG Parameters").grid(column=1,row=0,sticky=N)
        ttk.Label(view_frame,text="Camp").grid(column=0,row=2,sticky=N)
        ttk.Label(view_frame,text="EPICS").grid(column=1,row=2,sticky=N)
        
        self.text_nw.grid(column=0,row=1,sticky=(N,W,E,S))
        self.text_ne.grid(column=1,row=1,sticky=(N,W,E,S))
        self.text_sw.grid(column=0,row=3,sticky=(N,W,E,S))
        self.text_se.grid(column=1,row=3,sticky=(N,W,E,S))
        
        view_frame.grid(column=0,row=1)
        
        # details frame: stuff at the bottom ----------------------------------
        details_frame = ttk.Frame(file_tab)
        entry_rebin = Spinbox(details_frame,from_=1,to=100,width=3,\
                textvariable=self.rebin)
        
        # update check box
        self.is_updating = BooleanVar()
        self.is_updating.set(False)
        update_box = Checkbutton(details_frame,text='Periodic Redraw',
                command=self.do_update,variable=self.is_updating,onvalue=True,
                offvalue=False)

        # asymmetry type combobox
        self.asym_type = StringVar()
        self.asym_type.set('')
        self.entry_asym_type = ttk.Combobox(details_frame,\
                textvariable=self.asym_type,state='readonly',width=25)
        self.entry_asym_type['values'] = ()
                
        # gridding
        ttk.Label(details_frame,text="Rebin:").grid(column=0,row=0,sticky=E)
        entry_rebin.grid(column=1,row=0,sticky=E)
        self.entry_asym_type.grid(column=2,row=0,sticky=E)
        update_box.grid(column=3,row=0,sticky=E)
        details_frame.grid(column=0,row=2,sticky=N)
        
        # padding 
        for child in details_frame.winfo_children(): 
            child.grid_configure(padx=5, pady=5)
            
    # ======================================================================= #
    def __del__(self):
        pass
        
    # ======================================================================= #
    def draw(self):
        """Get data then draw."""
        if self.get_data():
            self.bfit.draw(self.data,\
                    self.asym_dict[self.asym_type.get()],rebin=self.rebin.get())
            
    # ======================================================================= #
    def export(self):
        """Export data as csv"""
        
        # get data
        if not self.get_data():
            return
        data = self.data
        
        # get filename 
        filename = filedialog.asksaveasfilename(
                initialfile=self.default_export_filename%(data.year,data.run))
        
        # write to file
        self.bfit.export(data,filename)
    
    # ======================================================================= #
    def get_data(self):
        """Display data and send bdata object to bfit draw list. 
        Return True on success, false on Failure
        """
        
        # settings
        mode_dict = {"20":"SLR","1f":"Frequency Scan","1n":"Rb Cell Scan",
                    '2h':'SLR with Alpha Tracking','2s':'Spin Echo',
                    '2e':'Randomized Frequency Scan'}
        
        # fetch data file
        try:
            year = int(self.year.get())
            run = int(self.runn.get())
        except ValueError:
            self.set_textbox_text(self.text_nw,'Input must be integer valued')
            self.set_textbox_text(self.text_ne,'Input must be integer valued')
            self.set_textbox_text(self.text_sw,'Input must be integer valued')
            self.set_textbox_text(self.text_se,'Input must be integer valued')
            return False
        
        try: 
            data = bdata(run,year=year)
        except ValueError:
            self.set_textbox_text(self.text_nw,'File read failed')
            self.set_textbox_text(self.text_ne,'File read failed')
            self.set_textbox_text(self.text_sw,'File read failed')
            self.set_textbox_text(self.text_se,'File read failed')
            return False
        except RuntimeError:
            self.set_textbox_text(self.text_nw,'File does not exist.')
            self.set_textbox_text(self.text_ne,'File does not exist.')
            self.set_textbox_text(self.text_sw,'File does not exist.')
            self.set_textbox_text(self.text_se,'File does not exist.')
            return False
        
        # set draw parameters
        self.bfit.set_asym_calc_mode_box(data.mode)
        
        # NE -----------------------------------------------------------------
        
        # get data: headers
        mode = mode_dict[data.mode]
        try:
            if data.ppg.rf_enable.mean and data.mode == '20':
                mode = "Hole Burning"
        except AttributeError:
            pass
        
        mins,sec = divmod(data.duration, 60)
        duration = "%dm %ds" % (mins,sec)
        
        # set dictionary
        data_nw =  {"Run":str(data.run),
                    "Area": data.area,
                    "Run Mode": "%s (%s)" % (mode,data.mode),
                    "Title": data.title,
                    "Experimenters": data.experimenter,
                    "Sample": data.sample,
                    "Orientation":data.orientation,
                    "Experiment":str(data.exp),
                    "Run Duration": duration,
                    "Start": data.start_date,
                    "End": data.end_date,
                    "":"",
                    }
        
        # set key order 
        key_order_nw = ['Run','Run Mode','Title','',
                        'Start','End','Run Duration','',
                        'Sample','Orientation','',
                        'Experiment','Area','Experimenters',
                        ]
        
        # SW -----------------------------------------------------------------
        data_sw = {'':''}
        key_order_sw = []
                        
        # get data: temperature and fields
        try:
            temp = data.camp.smpl_read_A.mean
            temp_stdv = data.camp.smpl_read_A.std
            data_sw["Temperature"] = "%.2f +/- %.2f K" % (temp,temp_stdv)
            key_order_sw.append('Temperature')
        except AttributeError:
            pass
        
        try:
            curr = data.camp.smpl_current
            data_sw["Heater Current"] = "%.2f +/- %.2f A" % (curr.mean,curr.std)
            key_order_sw.append('Heater Current')
        except AttributeError:
            pass
        
        try:
            temp = data.camp.oven_readC.mean
            temp_stdv = data.camp.oven_readC.std
            data_sw['Oven Temperature'] = "%.2f +/- %.2f K" % (temp,temp_stdv)
            key_order_sw.append('Oven Temperature')
        except AttributeError:
            pass
        
        try:
            curr = data.camp.oven_current
            data_sw['Oven Current'] = "%.2f +/- %.2f A" % (curr.mean,curr.std)
            key_order_sw.append('Oven Current')
        except AttributeError:
            pass
        
        try: 
            field = np.around(data.camp.b_field.mean,3)
            data_sw['Magnetic Field'] = "%.3f T" % field
            key_order_sw.append('Magnetic Field')
        except AttributeError:
            pass
            
        key_order_sw.append('')
                
        # cryo options
        try: 
            mass = data.camp.mass_read
            data_sw['Mass Flow'] = "%.3f +\- %.3f" % (mass.mean,mass.std)
            key_order_sw.append('Mass Flow')
        except AttributeError:
            pass
    
        try: 
            cryo = data.camp.cryo_read
            data_sw['CryoEx Mass Flow'] = "%.3f +\- %.3f" % (cryo.mean,cryo.std)
            key_order_sw.append('CryoEx Mass Flow')
        except AttributeError:
            pass    
            
        try: 
            data_sw['Needle Setpoint'] = "%.3f turns" % data.camp.needle_set.mean
            key_order_sw.append('Needle Setpoint')
        except AttributeError:
            pass    
            
        try: 
            data_sw['Needle Readback'] = "%.3f turns" % data.camp.needle_pos.mean
            key_order_sw.append('Needle Readback')
        except AttributeError:
            pass    
            
        try:
            lift_set = np.around(data.camp.clift_set.mean,3)
            data_sw['Cryo Lift Setpoint'] = "%.3f mm" % lift_set
            key_order_sw.append('Cryo Lift Setpoint')
        except AttributeError:
            pass
        
        try:
            lift_read = np.around(data.camp.clift_read.mean,3)
            data_sw['Cryo Lift Readback'] = "%.3f mm" % lift_read
            key_order_sw.append('Cryo Lift Readback')
        except AttributeError:
            pass
            
        # rf dac
        if mode != 'SLR':
            key_order_sw.append('')
            try: 
                data_sw['rf_dac'] = "%d" % int(data.camp.rf_dac.mean)
                key_order_sw.append('rf_dac')
            except AttributeError:
                pass
            
            try: 
                data_sw['RF Amplifier Gain'] = "%.2f" % data.camp.rfamp_rfgain.mean
                key_order_sw.append('RF Amplifier Gain')
            except AttributeError:
                pass    
                
                
        # SE -----------------------------------------------------------------
        data_se = {'':''}
        key_order_se = []
            
        # get data: biases 
        try:
            if 'nqr_bias' in data.epics.keys():
                bias =  data.epics.nqr_bias.mean/1000.
            elif 'nmr_bias' in data.epics.keys():
                bias =  data.epics.nmr_bias.mean
            
            data_se["Platform Bias"] = "%.3f kV" % np.around(bias,3)
            key_order_se.append("Platform Bias")
            
        except UnboundLocalError:
            pass
        
        try:
            data_se["BIAS15"] = "%.3f V" % np.around(data.epics.bias15.mean,3)
            key_order_se.append('BIAS15')
        except AttributeError:
            pass
        
        # get data: beam energy
        try: 
            init_bias = data.epics.target_bias.mean
        except AttributeError:
            try:
                init_bias = data.epics.target_bias.mean
            except AttributeError:
                pass
            
        try:
            val = np.around(init_bias/1000.,3)
            data_se["Initial Beam Energy"] = "%.3f keV" % val
            key_order_se.append('Initial Beam Energy')
        except UnboundLocalError:
            pass
        
        # Get final beam energy
        try: 
            val = np.around(data.beam_kev(),3)
            data_se['Implantation Energy'] = "%.3f keV" % val
            key_order_se.append('Implantation Energy')
        except AttributeError:
            pass
        
        key_order_se.append('')
        
        # laser stuff
        try: 
            val = data.epics.las_pwr
            data_se['Laser Power'] = "%.3f +\- %.3f A" % (val.mean,val.std)
            key_order_se.append('Laser Power')
        except AttributeError:
            pass
        
        # magnet stuff
        try: 
            val = data.epics.hh_current.mean
            data_se['Magnet Current'] = "%.3f A" % val
            key_order_se.append('Magnet Current')
            
            val = current2field(val)
            data_se['Magnetic Field'] = "%.3f Gauss" % val
            key_order_se.append('Magnetic Field')
        except AttributeError:
            pass
        
        # NE -----------------------------------------------------------------
        data_ne = {'':''}
        key_order_ne = []
        
        # get data: SLR data
        if data.mode in ['20','2h']:
            try:
                dwell = int(data.ppg.dwelltime.mean)
                data_ne['Dwell Time'] = "%d ms" % dwell
                key_order_ne.append('Dwell Time')
            except AttributeError:
                pass
            
            try:    
                beam = int(data.ppg.prebeam.mean)            
                data_ne['Number of Prebeam Dwelltimes'] = "%d dwelltimes" % beam
                key_order_ne.append('Number of Prebeam Dwelltimes')
            except AttributeError:
                pass
            
            try:    
                beam = int(data.ppg.beam_on.mean)            
                data_ne['Number of Beam On Dwelltimes'] = "%d dwelltimes" % beam
                key_order_ne.append('Number of Beam On Dwelltimes')
            except AttributeError:
                pass
            
            try: 
                beam = int(data.ppg.beam_off.mean)
                data_ne['Number of Beam Off Dwelltimes'] = "%d dwelltimes" % beam
                key_order_ne.append('Number of Beam Off Dwelltimes')
            except AttributeError:
                pass
            
            try:    
                rf = int(data.ppg.rf_on_delay.mean)
                data_ne['RF On Delay'] = "%d dwelltimes" % rf
                key_order_ne.append('RF On Delay')
            except AttributeError:
                pass
            
            try:    
                rf = int(data.ppg.rf_on.mean)
                data_ne['RF On Duration'] = "%d dwelltimes" % rf
                key_order_ne.append('RF On Duration')
            except AttributeError:
                pass
            
            try:    
                hel = bool(data.ppg.hel_enable.mean)
                data_ne['Flip Helicity'] = str(hel)
                key_order_ne.append('Flip Helicity')
            except AttributeError:
                pass
            
            try:    
                hel = int(data.ppg.hel_sleep.mean)
                data_ne['Helicity Flip Sleep'] = "%d ms" % hel
                key_order_ne.append('Helicity Flip Sleep')
            except AttributeError:
                pass
        
            key_order_ne.append('')
            
            try:
                rf = bool(data.ppg.rf_enable.mean)
                data_ne['RF Enable'] = str(rf)
                key_order_ne.append('RF Enable')
                
                if rf:
                    freq = int(data.ppg.freq.mean)    
                    data_ne['Frequency'] = "%d Hz" % freq
                    key_order_ne.append('Frequency')
            except AttributeError:
                pass
            
        # get 1F specific data
        elif data.mode in ['1f']:
            try:
                val = int(data.ppg.dwelltime.mean)
                data_ne['Bin Width'] = "%d ms" % val
                key_order_ne.append('Bin Width')
            except AttributeError:
                pass
            
            try:    
                val = int(data.ppg.nbins.mean)
                data_ne['Number of Bins'] = "%d" % val
                key_order_ne.append('Number of Bins')
            except AttributeError:
                pass
            
            try:
                val = bool(data.ppg.const_t_btwn_cycl.mean)
                data_ne['Enable Const Time Between Cycles'] = str(val)
                key_order_ne.append('Enable Const Time Between Cycles')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.freq_start.mean)
                data_ne['Frequency Scan Start'] = '%d Hz' % val
                key_order_ne.append('Frequency Scan Start')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.freq_stop.mean)
                data_ne['Frequency Scan End'] = '%d Hz' % val
                key_order_ne.append('Frequency Scan End')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.freq_incr.mean)
                data_ne['Frequency Scan Increment'] = '%d Hz' % val
                key_order_ne.append('Frequency Scan Increment')
            except AttributeError:
                pass
            
            try:
                val = bool(data.ppg.hel_enable.mean)
                data_ne['Flip Helicity'] = str(val)
                key_order_ne.append('Flip Helicity')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.hel_sleep.mean)
                data_ne['Helicity Flip Sleep'] = "%d ms" % val
                key_order_ne.append('Helicity Flip Sleep')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.ncycles.mean)
                data_ne['Number of Cycles per Scan Increment'] = '%d' % val
                key_order_ne.append('Number of Cycles per Scan Increment')
            except AttributeError:
                pass
            
        # get Rb Cell specific data
        elif data.mode in ['1n']:
            
            try:
                dwell = int(data.ppg.dwelltime.mean)
                data_ne['Bin Width'] = "%d ms" % dwell
                key_order_ne.append('Bin Width')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.volt_start.mean)
                data_ne['Start Rb Scan'] = '%d Volts' % val
                key_order_ne.append('Start Rb Scan')
            except AttributeError:
                pass
            
            try:    
                val = int(data.ppg.volt_stop.mean)
                data_ne['Stop Rb Scan'] = '%d Volts' % val
                key_order_ne.append('Stop Rb Scan')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.volt_incr.mean)
                data_ne['Scan Increment'] = '%d Volts' % val
                key_order_ne.append('Scan Increment')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.nbins.mean)
                data_ne['Number of Bins'] = '%d' % val
                key_order_ne.append('Number of Bins')
            except AttributeError:
                pass
            
            try:
                val = bool(data.ppg.hel_enable.mean)
                data_ne['Flip Helicity'] = str(val)
                key_order_ne.append('Flip Helicity')
            except AttributeError:
                pass
            
            try:
                val = int(data.ppg.hel_sleep.mean)
                data_ne['Helicity Flip Sleep'] = "%d ms" % val
                key_order_ne.append('Helicity Flip Sleep')
            except AttributeError:
                pass
        
        # get 2e mode specific data
        elif data.mode in ['2e']:
            
            try:
                val = int(data.ppg.rf_on_ms.mean)
                data_ne['RF On Time'] = "%d ms" % val
                key_order_ne.append('RF On Time')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.rf_on_delay.mean)
                data_ne['Number of RF On Delays'] = "%d" % val
                key_order_ne.append('Number of RF On Delays')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.beam_off_ms.mean)
                data_ne['Beam Off Time'] = "%d ms" % val
                key_order_ne.append('Beam Off Time')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.ndwell_post_on.mean)
                data_ne['Number of post RF BeamOn Dwelltimes'] = "%d" % val
                key_order_ne.append('Number of post RF BeamOn Dwelltimes')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.ndwell_per_f.mean)
                data_ne['Number of Dwelltimes per Frequency'] = "%d" % val
                key_order_ne.append('Number of Dwelltimes per Frequency')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.freq_start.mean)
                data_ne['Frequency Scan Start'] = "%d Hz" % val
                key_order_ne.append('Frequency Scan Start')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.freq_stop.mean)
                data_ne['Frequency Scan Stop'] = "%d Hz" % val
                key_order_ne.append('Frequency Scan Stop')
            except AttributeError:
                pass
                
            try:
                val = int(data.ppg.freq_incr.mean)
                data_ne['Frequency Scan Increment'] = "%d Hz" % val
                key_order_ne.append('Frequency Scan Increment')
            except AttributeError:
                pass
                
            try:
                val = bool(data.ppg.rand_freq_val.mean)
                data_ne['Randomize Frequency Scan Increments'] = str(val)
                key_order_ne.append('Randomize Frequency Scan Increments')
            except AttributeError:
                pass
                
            try:
                val = bool(data.ppg.hel_enable.mean)
                data_ne['Flip Helicity'] = str(val)
                key_order_ne.append('Flip Helicity')
            except AttributeError:
                pass
                
            try:
                val = bool(data.ppg.hel_enable.mean)
                data_ne['Helicity Flip Sleep'] = "%d ms" % val
                key_order_ne.append('Helicity Flip Sleep')
            except AttributeError:
                pass
                
            key_order_ne.append('')
            
        
        # set viewer string
        def set_str(data_dict,key_order,txtbox):
        
            m = max(max(map(len, list(data_dict.keys()))) + 1,5)
            s = '\n'.join([k.rjust(m)+': ' + data_dict[k] for k in key_order])
            self.set_textbox_text(txtbox,s)
        
        set_str(data_nw,key_order_nw,self.text_nw)
        set_str(data_ne,key_order_ne,self.text_ne)
        set_str(data_sw,key_order_sw,self.text_sw)
        set_str(data_se,key_order_se,self.text_se)
        
        # set data field
        self.data = data
        
        return True
   
    # ======================================================================= #
    def set_textbox_text(self,textbox,text):
        """Set the text in a tkinter Text widget"""
        textbox.delete('1.0',END)
        textbox.insert('1.0',text)
        
    # ======================================================================= #
    def do_update(self):
        
        if self.is_updating.get():
            self.draw()
            print('\rLast update:',str(datetime.datetime.now()).split('.')[0],
                  end='',flush=True)
            self.bfit.root.after(self.bfit.update_period*1000,self.do_update)
        
# =========================================================================== #




# fetch_files tab for bfit
# Derek Fujimoto
# Nov 2017

from tkinter import *
from tkinter import ttk, messagebox, filedialog
import numpy as np
import pandas as pd
from bdata import bdata
import datetime
from functools import partial
import matplotlib.pyplot as plt
from bfit.gui.fitdata import fitdata

__doc__="""
    To-do:
        scrollbar for lots of runs selected
    """

# =========================================================================== #
# =========================================================================== #
class fetch_files(object):
    """
        Data fields:
            
            check_rebin: IntVar for handling rebin aspect of checkall
            check_bin_remove: StringVar for handing omission of 1F data
            check_state: BooleanVar for handling check all
            entry_asym_type: combobox for asym calc and draw type
            year: StringVar of year to fetch runs from 
            run: StringVar input to fetch runs.
            data: pointer to bfit.data
            bfit: pointer to parent class
            data_lines: dictionary of dataline obj, keyed by run number
            fet_entry_frame: frame of fetch tab
            runmode_label: display run mode
            runmode: display run mode string
    """
    
    runmode_relabel = {'20':'Spin-Lattice Relaxation (20)',
                       '1f':'Frequency Scan (1f)',
                       '2e':'Randomized Frequency (2e)',
                       '1n':'Rb Cell Scan (1n)',
                       '2h':'Alpha Tagging/Diffusion (2h)'}
    run_number_starter_line = '40001 40005-40010 (run numbers)'
    bin_remove_starter_line = '1 5 100-200 (omit bins)'
    
    # ======================================================================= #
    def __init__(self,fetch_data_tab,bfit):
        
        # initialize
        self.bfit = bfit
        self.data_lines = {}
        self.fit_input_tabs = {}
        self.check_rebin = IntVar()
        self.check_bin_remove = StringVar()
        self.check_state = BooleanVar()
        self.fetch_data_tab = fetch_data_tab
        self.data = self.bfit.data
        
        # Frame for specifying files -----------------------------------------
        fet_entry_frame = ttk.Labelframe(fetch_data_tab,text='Specify Files')
        self.year = StringVar()
        self.run = StringVar()
        
        self.year.set(datetime.datetime.now().year)
        
        entry_year = ttk.Entry(fet_entry_frame,textvariable=self.year,width=5)
        entry_run = ttk.Entry(fet_entry_frame,textvariable=self.run,width=60)
        entry_run.insert(0,self.run_number_starter_line)
        entry_fn = partial(on_entry_click,text=self.run_number_starter_line,\
                            entry=entry_run)
        on_focusout_fn = partial(on_focusout,text=self.run_number_starter_line,\
                            entry=entry_run)
        entry_run.bind('<FocusIn>', entry_fn)
        entry_run.bind('<FocusOut>', on_focusout_fn)
        entry_run.config(foreground='grey')
        
        # fetch button
        fetch = ttk.Button(fet_entry_frame,text='Fetch',command=self.get_data)
        
        # grid and labels
        fet_entry_frame.grid(column=0,row=0,sticky=(N,E,W))
        ttk.Label(fet_entry_frame,text="Year:").grid(column=0,row=0,\
                sticky=(E))
        entry_year.grid(column=1,row=0,sticky=(E))
        ttk.Label(fet_entry_frame,text="Run Number:").grid(column=2,row=0,\
                sticky=(E))
        entry_run.grid(column=3,row=0,sticky=(E))
        fetch.grid(column=4,row=0,sticky=(E))
        
        # padding 
        for child in fet_entry_frame.winfo_children(): 
            child.grid_configure(padx=5, pady=5)
        
        # Frame for run mode -------------------------------------------------
        runmode_label_frame = ttk.Labelframe(fetch_data_tab,pad=(10,5,10,5),\
                text='Run Mode',)
        
        self.runmode_label = ttk.Label(runmode_label_frame,text="",font='bold',justify=CENTER)
        
        # Frame to hold datalines
        dataline_frame = ttk.Frame(fetch_data_tab,pad=5)
        
        # Frame to hold everything on the right ------------------------------
        bigright_frame = ttk.Frame(fetch_data_tab,pad=5)
        
        # Frame for group set options ----------------------------------------
        right_frame = ttk.Labelframe(bigright_frame,\
                text='Operations on Checked Items',pad=5)
        
        check_remove = ttk.Button(right_frame,text='Remove',\
                command=self.remove_all,pad=5)
        check_draw = ttk.Button(right_frame,text='Draw Data',\
                command=self.draw_all,pad=5)
        check_draw_fits = ttk.Button(right_frame,text='Draw Fits',\
                command=self.draw_all_fits,pad=5)
        
        check_set = ttk.Button(right_frame,text='Set',\
                command=self.set_all)
        check_rebin_label = ttk.Label(right_frame,text="SLR Rebin:",pad=5)
        check_rebin_box = Spinbox(right_frame,from_=1,to=100,width=3,\
                textvariable=self.check_rebin)
        check_bin_remove_entry = ttk.Entry(right_frame,\
                textvariable=self.check_bin_remove,width=20)
        
        check_all_box = ttk.Checkbutton(right_frame,
                text='Force Check State',variable=self.check_state,
                onvalue=True,offvalue=False,pad=5,command=self.check_all)
        self.check_state.set(False)
                
        check_toggle_button = ttk.Button(right_frame,\
                text='Toggle All Check States',command=self.toggle_all,pad=5)
        
        # add grey to check_bin_remove_entry
        check_bin_remove_entry.insert(0,self.bin_remove_starter_line)
        
        check_entry_fn = partial(on_entry_click,\
                text=self.bin_remove_starter_line,\
                entry=check_bin_remove_entry)
        
        check_on_focusout_fn = partial(on_focusout,\
                text=self.bin_remove_starter_line,\
                entry=check_bin_remove_entry)
        
        check_bin_remove_entry.bind('<FocusIn>', check_entry_fn)
        check_bin_remove_entry.bind('<FocusOut>', check_on_focusout_fn)
        check_bin_remove_entry.config(foreground='grey')
                
        # grid
        runmode_label_frame.grid(column=1,row=0,sticky=(N,W,E))
        self.runmode_label.grid(column=0,row=0,sticky=(N,W,E))
        
        bigright_frame.grid(column=1,row=1,sticky=(E,N,W))
        dataline_frame.grid(column=0,row=1,sticky=(E,W,S,N))
        
        right_frame.grid(column=0,row=0,sticky=(N,E,W))
        r = 0
        check_all_box.grid(         column=0,row=r,sticky=(N)); r+= 1
        check_toggle_button.grid(   column=0,row=r,sticky=(N),pady=10); r+= 1
        check_draw.grid(            column=0,row=r,sticky=(N))
        check_draw_fits.grid(       column=1,row=r,sticky=(N)); r+= 1
        check_remove.grid(          column=0,row=r,sticky=(N,E,W)); r+= 1
        check_rebin_label.grid(     column=0,row=r)
        check_rebin_box.grid(       column=1,row=r); r+= 1
        check_bin_remove_entry.grid(column=0,row=r,sticky=(N)); r+= 1
        check_set.grid(             column=0,row=r,sticky=(N))
        
        bigright_frame.grid(rowspan=20)
        check_all_box.grid(columnspan=2)
        check_remove.grid(columnspan=2)
        check_toggle_button.grid(columnspan=2)
        check_bin_remove_entry.grid(columnspan=2)
        check_set.grid(columnspan=2)
        
        check_rebin_box.grid_configure(padx=5,pady=5)
        check_rebin_label.grid_configure(padx=5,pady=5)
        check_set.grid_configure(padx=5,pady=5)
        
        # drawing style
        style_frame = ttk.Labelframe(bigright_frame,text='Drawing Quantity',\
                pad=5)
        self.entry_asym_type = ttk.Combobox(style_frame,\
                textvariable=self.bfit.fileviewer.asym_type,state='readonly',\
                width=20)
        self.entry_asym_type['values'] = ()
        
        style_frame.grid(column=0,row=1,sticky=(W,N))
        self.entry_asym_type.grid(column=0,row=0,sticky=(N))
        self.entry_asym_type.grid_configure(padx=24)
        
        # passing
        self.entry_run = entry_run
        self.entry_year = entry_year
        self.check_rebin_box = check_rebin_box
        self.check_bin_remove_entry = check_bin_remove_entry
        self.check_all_box = check_all_box
        self.dataline_frame = dataline_frame
    
    # ======================================================================= #
    def check_all(self):
        """Force all tickboxes to be in a given state"""
        for k in self.data_lines.keys():
            state = self.check_state.get()
            self.data_lines[k].check_state.set(state)
        
    # ======================================================================= #
    def draw_all(self,ignore_check=False):
        
        # condense drawing into a funtion
        def draw_lines():
            for r in self.data_lines.keys():
                if self.data_lines[r].check_state.get() or ignore_check:
                    self.data_lines[r].draw()
                
        # get draw style
        style = self.bfit.draw_style.get()
        
        # make new figure, draw stacked
        if style == 'stack':
            draw_lines()
            
        # overdraw in current figure, stacked
        elif style == 'redraw':
            plt.clf()
            self.bfit.draw_style.set('stack')
            draw_lines()
            self.bfit.draw_style.set('redraw')
            
        # make new figure, draw single
        elif style == 'new':
            plt.figure()
            self.bfit.draw_style.set('stack')
            draw_lines()
            self.bfit.draw_style.set('new')
        else:
            messagebox.showerror(message="Draw style not recognized")
            raise ValueError("Draw style not recognized")

    # ======================================================================= #
    def draw_all_fits(self,ignore_check=False):
        
        # condense drawing into a funtion
        def draw_lines():
            for r in self.data_lines.keys():
                line = self.data_lines[r] 
                if line.check_state.get() or ignore_check:
                    try:
                        self.bfit.fit_files.draw_fit(r,label=line.label.get())
                    except KeyError:
                        pass
                
        # get draw style
        style = self.bfit.draw_style.get()
        
        # make new figure, draw stacked
        if style == 'new':
            plt.figure()
            self.bfit.draw_style.set('stack')
            draw_lines()
            self.bfit.draw_style.set('new')
            
        elif style == 'stack':
            draw_lines()
            
        # overdraw in current figure, stacked
        elif style == 'redraw':
            plt.clf()
            self.bfit.draw_style.set('stack')
            draw_lines()
            self.bfit.draw_style.set('redraw')
            
        else:
            messagebox.showerror(message="Draw style not recognized")
            raise ValueError("Draw style not recognized")

    # ======================================================================= #
    def export(self):
        """Export all data files as csv"""
        
        # filename
        filename = self.bfit.fileviewer.default_export_filename
        try:
            filename = filedialog.askdirectory()+'/'+filename
        except TypeError:
            pass
        
        # get data and write
        for k in self.data.keys():
            d = self.data[k]
            self.bfit.export(d,filename%(d.year,d.run))
    
    # ======================================================================= #
    def get_data(self):
        """Split data into parts, and assign to dictionary."""
        
        # make list of run numbers, replace possible deliminators
        try:
            run_numbers = self.string2run(self.run.get())
        except ValueError:
            return
        
        # get data
        data = {}
        s = ['Failed to open run']
        for r in run_numbers:
            try:
                data[r] = fitdata(self.bfit,bdata(r,year=int(self.year.get())))
            except (RuntimeError,ValueError):
                s.append("%d (%d)" % (r,int(self.year.get())))

        # print error message
        if len(s)>1:
            s = '\n'.join(s)
            print(s)
            messagebox.showinfo(message=s)
        
        # check that data is all the same runtype
        run_types = []
        for k in self.data.keys():
            run_types.append(self.data[k].mode)
        for k in data.keys():
            run_types.append(data[k].mode)
            
        # different run types: select all runs of same type
        if not all([r==run_types[0] for r in run_types]):
            
            # unique run modes
            run_type_unique = np.unique(run_types)
            
            # message
            message = "Multiple run types detected:\n("
            for m in run_type_unique: 
                message += m+', '
            message = message[:-2]
            message += ')\n\nSelecting ' + run_types[0] + ' runs.'
            messagebox.showinfo(message=message)
            
        # get only run_types[0]
        for k in data.keys():
            if data[k].mode == run_types[0]:
                self.data[k] = data[k]
        
        try:
            self.runmode = run_types[0]
        except IndexError:
            messagebox.showerror(message='No valid runs detected.')
            raise RuntimeError('No valid runs detected.')
        self.runmode_label['text'] = self.runmode_relabel[self.runmode]
        self.bfit.set_asym_calc_mode_box(self.runmode)
        
        keys_list = list(self.data.keys())
        keys_list.sort()
        
        # make lines
        n = 1
        for r in keys_list:
            if r in self.data_lines.keys():
                self.data_lines[r].grid(n)
            else:
                self.data_lines[r] = dataline(self.bfit,\
                        self.data_lines,self.dataline_frame,self.data[r],n)
            n+=1
            
        self.bfit.fit_files.populate()
        
    # ======================================================================= #
    def remove_all(self):
        """Remove all data files from self.data_lines"""
        del_list = []
        for r in self.data_lines.keys():
            if self.data_lines[r].check_state.get():
                del_list.append(self.data_lines[r])
        for d in del_list:
            d.remove()
    
    # ======================================================================= #
    def return_binder(self):
        """Switch between various functions of the enter button. """
        
        # check where the focus is
        focus_id = str(self.bfit.root.focus_get())
        
        # run or year entry
        if focus_id in [str(self.entry_run), str(self.entry_year)]:
            self.get_data()
        
        # checked rebin or checked run omission
        elif focus_id in [str(self.check_rebin_box),\
                          str(self.check_bin_remove_entry)]:
            self.set_all()
        elif focus_id == str(self.check_all_box):
            self.draw_all()
        else:
            pass

    # ======================================================================= #
    def set_all(self):
        """Set a particular property for all checked items. """
        
        # check all file lines
        for r in self.data_lines.keys():
            
            # if checked
            if self.data_lines[r].check_state.get():
                
                # get values to enter
                self.data_lines[r].rebin.set(self.check_rebin.get())
                new_text = self.check_bin_remove.get()
                
                # check for greyed text
                if new_text != self.bin_remove_starter_line:
                    self.data_lines[r].bin_remove.set(new_text)
                else:
                    self.data_lines[r].bin_remove.set("")
                    
                # generate focus out event: trigger grey text reset
                self.data_lines[r].bin_remove_entry.event_generate('<FocusOut>')

    # ======================================================================= #
    def string2run(self,string):
        """Parse string, return list of run numbers"""
        
        full_string = string.replace(',',' ').replace(';',' ')
        full_string = full_string.replace(':','-')
        part_string = full_string.split()
        
        run_numbers = []
        for s in part_string:
            if '-' in s:
                try:
                    rn_lims = [int(s2) for s2 in s.split('-')]
                except ValueError:
                    run_numbers.append(int(s.replace('-','')))
                else:
                    rns = np.arange(rn_lims[0],rn_lims[1]+1).tolist()
                    run_numbers.extend(rns)
            else:
                run_numbers.append(int(s))
        # sort
        run_numbers.sort()
        
        if len(run_numbers) > 50:
            raise RuntimeWarning("Too many files selected (max 50).")
        return run_numbers
    
    # ======================================================================= #
    def toggle_all(self):
        """Toggle all tickboxes"""
        for k in self.data_lines.keys():
            state = not self.data_lines[k].check_state.get()
            self.data_lines[k].check_state.set(state)

        
# =========================================================================== #
# =========================================================================== #
class dataline(object):
    """
        A line of objects to display run properties and remove bins and whatnot.
        
        bfit:           pointer to root 
        bin_remove:     StringVar for specifying which bins to remove in 1f runs
        bin_remove_entry: Entry object for bin remove 
        check_state:    BooleanVar for specifying check state
        group:          IntVar for fitting group ID
        label:          StringVar for labelling runs in legends
        line_frame:     Frame that object is placed in
        lines_list:     list of datalines
        mode:           bdata run mode
        rebin:          IntVar for SLR rebin
        row:            position in list
        run:            bdata run number
        year:           bdata year
        
        
    """
        
    bin_remove_starter_line = '1 5 100-200 (omit bins)'
    
    # ======================================================================= #
    def __init__(self,bfit,lines_list,fetch_tab_frame,bdfit,row):
        """
            Inputs:
                fetch_tab_frame: parent in which to place line
                bdfit: fitdata object corresponding to the file which is placed here. 
                row: where to grid this object
        """
        
        # variables
        self.bfit = bfit
        self.bin_remove = bdfit.omit
        self.label = bdfit.label
        self.rebin = bdfit.rebin
        self.group = bdfit.group
        self.check_state = bdfit.check_state
        self.mode = bdfit.mode
        self.run =  bdfit.run
        self.year = bdfit.year
        self.row = row
        self.lines_list = lines_list
        bd = bdfit.bd
        self.bdfit = bdfit
        
        # temperature
        try:
            self.temperature = int(np.round(bdfit.temperature))
        except AttributeError:
            self.temperature = -1
            
        # field
        self.field = np.around(bdfit.field,2)
        
        if self.field > 0:
            field_text = "%.2f T"%self.field
        else:
            field_text = ' '*6
        
        # bias
        self.bias = self.bdfit.bias
        if self.bias > 0:
            bias_text = "%.2f kV"%self.bias
        else:
            bias_text = ' '*7
        
        # build objects
        line_frame = ttk.Frame(fetch_tab_frame,pad=(5,0))
        year_label = ttk.Label(line_frame,text="%d"%self.year,pad=5)
        run_label = ttk.Label(line_frame,text="%d"%self.run,pad=5)
        temp_label = ttk.Label(line_frame,text="%3d K"%self.temperature,pad=5)
        field_label = ttk.Label(line_frame,text=field_text,pad=5)
        bias_label = ttk.Label(line_frame,text=bias_text,pad=5)
        bin_remove_entry = ttk.Entry(line_frame,textvariable=self.bin_remove,\
                width=20)
                
        label_label = ttk.Label(line_frame,text="Label:",pad=5)
        label_entry = ttk.Entry(line_frame,textvariable=self.label,\
                width=10)
                
        remove_button = ttk.Button(line_frame,text='Remove',\
                command=self.remove,pad=1)
        draw_button = ttk.Button(line_frame,text='Draw',command=self.draw,pad=1)
        self.draw_fit_button = ttk.Button(line_frame,text='Draw Fit',
                command= lambda : self.bfit.fit_files.draw_fit(run=self.run),
                pad=1,state=DISABLED)
        
        rebin_label = ttk.Label(line_frame,text="Rebin:",pad=5)
        rebin_box = Spinbox(line_frame,from_=1,to=100,width=3,\
                textvariable=self.rebin)
                
        group_label = ttk.Label(line_frame,text="Group:",pad=5)
        group_box = Spinbox(line_frame,from_=1,to=100,width=3,\
                textvariable=self.group)
                   
        self.check_state.set(False)
        check = ttk.Checkbutton(line_frame,text='',variable=self.check_state,\
                onvalue=True,offvalue=False,pad=5)
         
        # add grey text to bin removal
        bin_remove_entry.insert(0,self.bin_remove_starter_line)
        entry_fn = partial(on_entry_click,\
                text=self.bin_remove_starter_line,entry=bin_remove_entry)
        on_focusout_fn = partial(on_focusout,\
                text=self.bin_remove_starter_line,entry=bin_remove_entry)
        bin_remove_entry.bind('<FocusIn>', entry_fn)
        bin_remove_entry.bind('<FocusOut>', on_focusout_fn)
        bin_remove_entry.config(foreground='grey')
             
        # add grey text to label
        label_entry.insert(0,str(self.run))
        entry_fn_lab = partial(on_entry_click,text=str(self.run),
                               entry=label_entry)
        on_focusout_fn_lab = partial(on_focusout,text=str(self.run),
                                 entry=label_entry)
        label_entry.bind('<FocusIn>', entry_fn_lab)
        label_entry.bind('<FocusOut>', on_focusout_fn_lab)
        label_entry.config(foreground='grey')
                
        # grid
        c = 1
        check.grid(column=c,row=0,sticky=E); c+=1
        year_label.grid(column=c,row=0,sticky=E); c+=1
        run_label.grid(column=c,row=0,sticky=E); c+=1
        temp_label.grid(column=c,row=0,sticky=E); c+=1
        field_label.grid(column=c,row=0,sticky=E); c+=1
        bias_label.grid(column=c,row=0,sticky=E); c+=1
        if self.mode in ['1f','1n']: 
            bin_remove_entry.grid(column=c,row=0,sticky=E); c+=1
        if self.mode in ['20','2h']: 
            rebin_label.grid(column=c,row=0,sticky=E); c+=1
            rebin_box.grid(column=c,row=0,sticky=E); c+=1
        label_label.grid(column=c,row=0,sticky=E); c+=1
        label_entry.grid(column=c,row=0,sticky=E); c+=1
        group_label.grid(column=c,row=0,sticky=E); c+=1
        group_box.grid(column=c,row=0,sticky=E); c+=1
        draw_button.grid(column=c,row=0,sticky=E); c+=1
        self.draw_fit_button.grid(column=c,row=0,sticky=E); c+=1
        remove_button.grid(column=c,row=0,sticky=E); c+=1
        
        # passing
        self.line_frame = line_frame
        self.bin_remove_entry = bin_remove_entry
        
        # grid frame
        self.grid(row)
        
    # ======================================================================= #
    def grid(self,row):
        """Re-grid a dataline object so that it is in order by run number"""
        self.row = row
        self.line_frame.grid(column=0,row=row,columnspan=2, sticky=(W,N))
        
    # ======================================================================= #
    def remove(self):
        """Remove displayed dataline object from file selection. """
        
        # kill buttons and fram
        for child in self.line_frame.winfo_children():
            child.destroy()
        for child in self.line_frame.winfo_children():
            child.destroy()
        self.line_frame.destroy()
        
        # get rid of data
        del self.bfit.fetch_files.data[self.run]
        del self.lines_list[self.run]
        
        # repopulate fit files tab
        self.bfit.fit_files.populate()
        
        # remove data from storage
        if len(self.lines_list) == 0:
            ff = self.bfit.fetch_files
            ff.runmode_label['text'] = ''
                
    # ======================================================================= #
    def draw(self):
        """Draw single data file."""
        
        # get new data file
        data = bdata(self.run,year=self.year)
        
        # get data file run type
        d = self.bfit.fileviewer.asym_type.get()
        d = self.bfit.fileviewer.asym_dict[d]
        
        if self.bin_remove.get() == self.bin_remove_starter_line:
            self.bfit.draw(data,d,self.rebin.get(),
                label=self.label.get())
        else:
            self.bfit.draw(data,d,self.rebin.get(),\
                option=self.bin_remove.get(),label=self.label.get())
        
# =========================================================================== #
def on_entry_click(event,entry,text):
    """Vanish grey text on click"""
    if entry.get() == text:
        entry.delete(0, "end") # delete all the text in the entry
        entry.insert(0, '') #Insert blank for user input
        entry.config(foreground = 'black')

# =========================================================================== #
def on_focusout(event,entry,text):
    """Set grey text for boxes on exit"""
    if entry.get() == '':
        entry.insert(0,text)
        entry.config(foreground = 'grey')
    else:
        entry.config(foreground = 'black')






from bfit.fitting.continuous import fscan

try:
    from bfit.fitting.integrator import PulsedFns
except ImportError:
    print("\n integrator not compiled. From withing bfit/fitting/, call: "+\
          "'python3 setup_integrator.py build_ext --inplace'")
else:
    from bfit.fitting.pulsed import slr


from bfit.fitting.default_routines import fitter



# fit_files tab for bfit
# Derek Fujimoto
# Dec 2017

from tkinter import *
from tkinter import ttk, messagebox, filedialog
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import datetime, os, traceback

from functools import partial
from bfit.gui.zahersCalculator import current2field
from bdata import bdata

# =========================================================================== #
# =========================================================================== #
class fit_files(object):
    """
        Data fields:
            chi_threshold:  if chi > thres, set color to red
            data            pointer to bfit.data
            file_tabs:      dictionary of fitinputtab objects, keyed by group number 
            fitter:         fitting object from self.bfit.routine_mod
            fit_function_title: title of fit function to use
            fit_function_title_box: spinbox for fit function names
            fit_input:      fitting input values = (fn_name,ncomp,data_list)
            fit_output:     fitting results, output of fitter
            groups:         group numbers from fetch tab
            mode:           what type of run is this. 
            n_component:    number of fitting components (IntVar)
            runbook:        notebook of fit inputs for runs
            runmode_label:  display run mode 
            set_as_group:   BooleanVar() if true, set fit parfor whole group
            xaxis:          StringVar() for parameter to draw on x axis
            yaxis:          StringVar() for parameter to draw on y axis
            xaxis_combobox: box for choosing x axis draw parameter
            yaxis_combobox: box for choosing y axis draw parameter
    """ 
    
    default_fit_functions = {'20':('Exp','Str Exp'),'2h':('Exp','Str Exp'),
            '1f':('Lorentzian','Gaussian'),'1n':('Lorentzian','Gaussian')}
    mode = ""
    chi_threshold = 1.5 # threshold for red highlight on bad fits 
    n_fitx_pts = 500    # number of points to draw in fitted curves

    # define draw componeents in draw_param
    draw_components = ['Temperature (K)','B0 Field (T)', 'RF Level DAC', 
                       'Platform Bias (kV)', 'Impl. Energy (keV)', 
                       'Run Duration (s)', 'Run Number','Sample', 'Start Time']

    # ======================================================================= #
    def __init__(self,fit_data_tab,bfit):
        
        # initialize
        self.file_tabs = {}
        self.bfit = bfit
        self.data = self.bfit.data
        self.groups = []
        self.fit_output = {}
        self.fitter = self.bfit.routine_mod.fitter()
            
        # make top level frames
        top_fit_frame = ttk.Frame(fit_data_tab,pad=5)   # fn select, run mode
        mid_fit_frame = ttk.Frame(fit_data_tab,pad=5)   # notebook
        right_frame = ttk.Labelframe(fit_data_tab,text='Fit Results',pad=5)     # draw fit results
        
        top_fit_frame.grid(column=0,row=0,sticky=(N,W))
        mid_fit_frame.grid(column=0,row=1,sticky=(N,W))
        right_frame.grid(column=1,row=1,columnspan=2,rowspan=2,sticky=(N,W,E))
        
        # TOP FRAME 
        
        # fit function select 
        fn_select_frame = ttk.Labelframe(fit_data_tab,text='Fit Function')
        self.fit_function_title = StringVar()
        self.fit_function_title.set("")
        self.fit_function_title_box = ttk.Combobox(fn_select_frame, 
                textvariable=self.fit_function_title,state='readonly')
        self.fit_function_title.trace('w', self.populate_param)
        
        # number of components in fit spinbox
        self.n_component = IntVar()
        self.n_component.set(1)
        n_component_box = Spinbox(fn_select_frame,from_=1,to=20, 
                textvariable=self.n_component,width=5,command=self.populate_param)
        
        # fit button
        fit_button = ttk.Button(fn_select_frame,text='Fit',command=self.do_fit,\
                                pad=1)
        
        # set as group checkbox
        self.set_as_group = BooleanVar()
        set_group_check = ttk.Checkbutton(fn_select_frame,
                text='Set Parameters for Group',\
                variable=self.set_as_group,onvalue=True,offvalue=False)
        
        # run mode 
        fit_runmode_label_frame = ttk.Labelframe(fit_data_tab,pad=(10,5,10,5),
                text='Run Mode',)
        self.fit_runmode_label = ttk.Label(fit_runmode_label_frame,text="",
                font='bold',justify=CENTER)
        
        # fitting routine
        fit_routine_label_frame = ttk.Labelframe(fit_data_tab,pad=(10,5,10,5),
                text='Fitting Routine',)
        self.fit_routine_label = ttk.Label(fit_routine_label_frame,text="",
                font='bold',justify=CENTER)
                
        # GRIDDING
            
        # top frame gridding
        fn_select_frame.grid(column=0,row=0,sticky=(E,W))
        self.fit_function_title_box.grid(column=0,row=0)
        ttk.Label(fn_select_frame,text="Number of Components:").grid(column=1,
                row=0,sticky=(E),padx=5,pady=5)
        n_component_box.grid(column=2,row=0,padx=5,pady=5)
        fit_button.grid(column=3,row=0,padx=1,pady=1)
        set_group_check.grid(column=4,row=0,padx=1,pady=1)
        
        # run mode gridding
        fit_runmode_label_frame.grid(column=1,row=0,sticky=(E,W))
        self.fit_runmode_label.grid(column=0,row=0,sticky=(E,W))
        
        # routine label gridding
        fit_routine_label_frame.grid(column=2,row=0,sticky=(E,W))
        self.fit_routine_label.grid(column=0,row=0,sticky=(E,W))
        
        # MID FRAME        
        self.runbook = ttk.Notebook(mid_fit_frame)
        self.runbook.grid(column=0,row=0)
        
        # RIGHT FRAME
        
        # draw and export buttons
        button_frame = ttk.Frame(right_frame)
        draw_button = ttk.Button(button_frame,text='Draw',command=self.draw_param)
        export_button = ttk.Button(button_frame,text='Export',command=self.export)
        
        # menus for x and y values
        ttk.Label(right_frame,text="x axis:").grid(column=0,row=1)
        ttk.Label(right_frame,text="y axis:").grid(column=0,row=2)
        
        self.xaxis = StringVar()
        self.yaxis = StringVar()
        
        self.xaxis.set('')
        self.yaxis.set('')
        
        self.xaxis_combobox = ttk.Combobox(right_frame,textvariable=self.xaxis,
                                      state='readonly',width=15)
        self.yaxis_combobox = ttk.Combobox(right_frame,textvariable=self.yaxis,
                                      state='readonly',width=15)
        
        # gridding
        button_frame.grid(column=0,row=0,columnspan=2)
        draw_button.grid(column=0,row=0,padx=5,pady=5)
        export_button.grid(column=1,row=0,padx=5,pady=5)
        
        self.xaxis_combobox.grid(column=1,row=1,pady=5)
        self.yaxis_combobox.grid(column=1,row=2,pady=5)
        
    # ======================================================================= #
    def populate(self,*args):
        """
            Make tabs for setting fit input parameters. 
        """
        
        # get data
        self.data = self.bfit.data
        
        # get groups 
        dl = self.bfit.fetch_files.data_lines
        self.groups = np.unique([dl[k].group.get() for k in dl.keys()])
        
        # get run mode by looking at one of the data dictionary keys
        for key_zero in self.data.keys(): break

        # check if clearing of old tabs is needed
        keys = self.file_tabs.keys()
        do_create = not all((g in keys for g in self.groups))
        do_create = do_create or not all((g in self.groups for g in keys))
        
        # clear old tabs
        if do_create:
            for child in self.runbook.winfo_children():
                child.destroy()
        
            # reset fit function combobox options
            try:               
                if self.mode != self.data[key_zero].mode:
                    # set run mode 
                    self.mode = self.data[key_zero].mode 
                    self.fit_runmode_label['text'] = \
                            self.bfit.fetch_files.runmode_relabel[self.mode]
                    
                    # set routine
                    self.fit_routine_label['text'] = self.fitter.__name__
                    
                    # set run functions        
                    fn_titles = self.fitter.function_names[self.mode]
                    self.fit_function_title_box['values'] = fn_titles
                    self.fit_function_title.set(fn_titles[0])
                        
            except UnboundLocalError:
                self.fit_function_title_box['values'] = ()
                self.fit_function_title.set("")
                self.fit_runmode_label['text'] = ""
                self.mode = ""

                        
            # make fitinputtab objects, clean up old tabs
            for g in self.groups:
                
                # add to list of groups
                if not g in self.file_tabs.keys():
                    self.file_tabs[g] = fitinputtab(self.bfit,self.runbook,g)
                
            # clean up old tabs
            del_list = [k for k in self.file_tabs.keys() if not k in self.groups]
            
            for k in del_list:
                del self.file_tabs[k]
                        
            # add tabs to notebook
            for k in self.file_tabs.keys():
                self.file_tabs[k].create()
            
            # populate the list of parameters 
            self.populate_param()
    
        else:
            for k in self.file_tabs.keys():
                self.file_tabs[k].update()
        
    # ======================================================================= #
    def populate_param(self,*args):
        """Populate the list of parameters"""
        
        # populate tabs
        for k in self.file_tabs.keys():
            self.file_tabs[k].populate_param()
            
        # populate axis comboboxes
        lst = self.draw_components.copy()
        lst.sort()
        
        try:
            parlst = [p for p in self.fitter.gen_param_names(
                                                self.fit_function_title.get(),
                                                self.n_component.get())]
        except KeyError:
            self.xaxis_combobox['values'] = []
            self.yaxis_combobox['values'] = []
            return
            
        parlst.sort()
        
        self.xaxis_combobox['values'] = parlst+lst
        self.yaxis_combobox['values'] = parlst+lst
            
    # ======================================================================= #
    def do_fit(self,*args):
        
        # fitter
        fitter = self.fitter
        
        # get fitter inputs
        fn_name = self.fit_function_title.get()
        ncomp = self.n_component.get()
        
        # build data list
        data_list = []
        for g in self.groups:
            tab = self.file_tabs[g]
            collist = tab.collist
            runlist = tab.runlist
            
            for r in runlist:
                
                # bdata object
                bdfit = self.data[r]
                bdataobj = bdfit.bd
                
                # pdict
                pdict = {}
                for parname in tab.parentry.keys():
                    
                    # get entry values
                    pline = tab.parentry[parname]
                    line = []
                    for col in collist:
                        
                        # get number entries
                        if col in ['p0','blo','bhi']:
                            try:
                                line.append(float(pline[col][0].get()))
                            except ValueError as errmsg:
                                messagebox.showerror("Error",str(errmsg))
                        
                        # get "Fixed" entry
                        elif col in ['fixed']:
                            line.append(pline[col][0].get())
                    
                    # make dict
                    pdict[parname] = line
                    
                # doptions
                doptions = {}
                doptions['rebin'] = bdfit.rebin.get()
                doptions['group'] = bdfit.group.get()
                
                if self.mode == '1f':
                    dline = self.bfit.fetch_files.data_lines[r]
                    doptions['omit'] = dline.bin_remove.get()
                    if doptions['omit'] == dline.bin_remove_starter_line: 
                        doptions['omit'] = ''
                    
                elif self.mode == '20':
                    pass
                    
                elif self.mode == '2h':
                    pass
                    
                elif self.mode == '2e':
                    raise RuntimeError('2e fitting not implemented')
                
                else:
                    raise RuntimeError('Fitting mode not recognized')
                
                # make data list
                data_list.append([bdataobj,pdict,doptions])
        
        # call fitter with error message, potentially
        self.fit_input = (fn_name,ncomp,data_list)
        
        # make fitting status window
        fit_status_window = Toplevel(self.bfit.root)
        fit_status_window.lift()
        fit_status_window.resizable(FALSE,FALSE)
        ttk.Label(fit_status_window,text="Please Wait",pad=20).grid(column=0,
                                                    row=0,sticky=(N,S,E,W))
        fit_status_window.update_idletasks()
        self.bfit.root.update_idletasks()
        
        width = fit_status_window.winfo_reqwidth()
        height = fit_status_window.winfo_reqheight()
        
        rt_x = self.bfit.root.winfo_x()
        rt_y = self.bfit.root.winfo_y()
        rt_w = self.bfit.root.winfo_width()
        rt_h = self.bfit.root.winfo_height()
        
        x = rt_x + rt_w/2 - (width/2)
        y = rt_y + rt_h/3 - (width/2)
        
        fit_status_window.geometry('{}x{}+{}+{}'.format(width, height, int(x), int(y)))
        fit_status_window.update_idletasks()
        
        # do fit then kill window
        try:
            # fit_output keyed as {run:[key/par/cov/chi/fnpointer]}
            fit_output = fitter(fn_name=fn_name,ncomp=ncomp,
                                     data_list=data_list,
                                     hist_select=self.bfit.hist_select)
        except Exception as errmsg:
            fit_status_window.destroy()
            messagebox.showerror("Error",str(errmsg))
            raise errmsg
        else:
            fit_status_window.destroy()
        
        # set output results
        for run in fit_output.keys():
            self.data[run].set_fitresult(fit_output[run])
            
        # display run results
        for g in self.groups:
            self.file_tabs[g].set_display()
            self.file_tabs[g].set_run_color()
        
        # enable draw buttons on fetch files tab
        for r in runlist:
            self.bfit.fetch_files.data_lines[r].draw_fit_button['state'] = 'normal'
        
        # draw fit results
        self.bfit.fetch_files.draw_all(ignore_check=True)
        
        style = self.bfit.draw_style.get()
        
        if style in ['redraw','new']:
            self.bfit.draw_style.set('stack')
        
        self.bfit.fetch_files.draw_all_fits(ignore_check=True)
        self.bfit.draw_style.set(style)
            
    # ======================================================================= #
    def draw_fit(self,run,**drawargs):
        """Draw fit for a single run"""
        
        # Settings
        xlabel_dict={'20':"Time (s)",
                     '2h':"Time (s)",
                     '2e':'Frequency (MHz)',
                     '1f':'Frequency (MHz)',
                     '1n':'Voltage (V)'}
                     
        # get data and fit results
        data = self.data[run]
        fit_par = [data.fitpar['res'][p] for p in data.parnames]
        fn = data.fitfn
        data = data.bd
        
        # get draw style
        style = self.bfit.draw_style.get()
        
        # label reset
        if 'label' not in drawargs.keys():
            drawargs['label'] = self.data[run].label.get()+'_fit'
        elif 'fit' not in drawargs['label']:
            drawargs['label'] += ' (fit)'
        label = drawargs['label']
        
        # set drawing style
        if style == 'new':
            plt.figure()
        if style == 'stack':
            
            ax = plt.gca()
            try:
                idx = [ell.get_label() for ell in ax.lines].index(label)
            except ValueError as err:
                pass
            else:
                del ax.lines[idx]              # clear lines 
                
        elif style == 'redraw':
            ylim = ax.get_ylim()
            xlim = ax.get_xlim()
            plt.clf()
            plt.ylim(*ylim)
            plt.xlim(*xlim)
            
        # set drawing style arguments
        for k in self.bfit.style:
            if k not in drawargs.keys() \
                    and 'marker' not in k \
                    and k not in ['elinewidth','capsize']:
                drawargs[k] = self.bfit.style[k]
        
        # linestyle reset
        if drawargs['linestyle'] == 'None': 
            drawargs['linestyle'] = '-'
        
        # draw
        t,a,da = data.asym('c')
        fitx = np.arange(self.n_fitx_pts)/float(self.n_fitx_pts)*\
                                                    (max(t)-min(t))+min(t)
        
        if   data.mode == '1f': fitxx = fitx*self.bfit.freq_unit_conv
        elif data.mode == '1n': fitxx = fitx*self.bfit.volt_unit_conv
        else:                   fitxx = fitx
    
        plt.plot(fitxx,fn(fitx,*fit_par),zorder=10,**drawargs)
        
        # plot elements
        plt.ylabel('Asymmetry')
        plt.xlabel(xlabel_dict[self.mode])
        
        # show
        plt.tight_layout()
        plt.legend()
        
    # ======================================================================= #
    def draw_param(self,*args):
        
        # make sure plot shows
        plt.ion()
        
        # get draw components
        xdraw = self.xaxis.get()
        ydraw = self.yaxis.get()
        
        # get plottable data
        try:
            xvals, xerrs = self.get_values(xdraw)
            yvals, yerrs = self.get_values(ydraw)
        except UnboundLocalError as err:
            messagebox.showerror("Error",'Select two input parameters')
            raise err
        except (KeyError,AttributeError) as err:
            messagebox.showerror("Error",
                    'Drawing parameter "%s" or "%s" not found' % (xdraw,ydraw))
            raise err
            
        # get draw style
        style = self.bfit.draw_style.get()
        
        if style == 'new':
            plt.figure()
        elif style == 'redraw':
            plt.clf()
        plt.gca().get_xaxis().get_major_formatter().set_useOffset(False)
        plt.gca().get_yaxis().get_major_formatter().set_useOffset(False)
        
        # draw
        if type(xvals[0]) == str:
            plt.xticks(np.arange(len(xvals)))
            plt.gca().set_xticklabels(xvals)
            xvals = np.arange(len(xvals))
        
        if type(yvals[0]) == str:
            plt.yticks(np.arange(len(yvals)))
            plt.gca().set_yticklabels(yvals)
            yvals = np.arange(len(yvals))
        
        plt.errorbar(xvals,yvals,xerr=xerrs,yerr=yerrs,fmt='.')
            
        # plot elements
        plt.xlabel(xdraw)
        plt.ylabel(ydraw)
        plt.tight_layout()
        
    # ======================================================================= #
    def export(self):
        
        # get values and errors
        val = {}
        for v in self.xaxis_combobox['values']:
            try:
                v2 = self.get_values(v) 
            except AttributeError: 
                traceback.print_exc()
            else:
                val[v] = v2[0]
                val['Error '+v] = v2[1]
        
        # make data frame for output
        df = pd.DataFrame(val)
        df.set_index('Run Number',inplace=True)
        
        # get file name
        filename = filedialog.asksaveasfilename()
        
        # check extension 
        if os.path.splitext(filename)[1] == '':
            filename += '.csv'
        df.to_csv(filename)
        
    # ======================================================================= #
    def get_values(self,select):
        """ Get plottable values"""
        data = self.data
        runs = list(data.keys())
        runs.sort()
        parout = self.fit_output
    
        # Data file options
        if select == 'Temperature (K)':
            val = [data[r].camp.smpl_read_A.mean for r in runs]
            err = [data[r].camp.smpl_read_A.std for r in runs]
        
        elif select == 'B0 Field (T)':
            try:
                val = [data[r].camp.b_field.mean for r in runs]
                err = [data[r].camp.b_field.std for r in runs]
            except AttributeError:
                val = [current2field(data[r].epics.hh_current.mean)/1e4 for r in runs]
                err = [current2field(data[r].epics.hh_current.std)/1e4 for r in runs]
        
        elif select == 'RF Level DAC':
            val = [data[r].camp.rf_dac.mean for r in runs]
            err = [data[r].camp.rf_dac.std for r in runs]
        
        elif select == 'Platform Bias (kV)':
            try:
                val = [data[r].epics.nmr_bias.mean for r in runs]
                err = [data[r].epics.nmr_bias.std for r in runs]
            except AttributeError:
                val = [data[r].epics.nqr_bias.mean/1000. for r in runs]
                err = [data[r].epics.nqr_bias.std/1000. for r in runs]
                
        elif select == 'Impl. Energy (keV)':
            val =  [data[r].beam_kev() for r in runs]
            err =  [0 for r in runs]
        
        elif select == 'Run Duration (s)':
            val = [data[r].duration for r in runs]
            err = [0 for r in runs]
        
        elif select == 'Run Number':
            val = [data[r].run for r in runs]
            err = [0 for r in runs]
        
        elif select == 'Sample':
            val = [data[r].sample for r in runs]
            err = [0 for r in runs]
            
        elif select == 'Start Time':
            val = [data[r].start_date for r in runs]
            err = [0 for r in runs]
        
        # fitted parameter options
        elif select in self.fitter.gen_param_names(self.fit_function_title.get(),
                                                   self.n_component.get()):
            val = []
            err = []
            
            for r in runs:
                try:
                    val.append(parout[r][1][parout[r][0].index(select)])
                    err.append(parout[r][2][parout[r][0].index(select)])
                except KeyError:
                    val.append(np.nan)
                    err.append(np.nan)
        
        return (val,err)
    
# =========================================================================== #
# =========================================================================== #
class fitinputtab(object):
    """
        Instance variables 
        
            bfit        pointer to top class
            data        pointer to bfit.data
            parent      pointer to parent object (frame)
            group       fitting group number
            parlabels   label objects, saved for later destruction
            parentry    [parname][colname] of ttk.Entry objects saved for 
                            retrieval and destruction
            runbox      listbox with run numbers: select which result to display
            runlist     list of run numbers to fit
            selected    index of selected run in runbox (int)
            fitframe    mainframe for this tab. 
    """
    
    n_runs_max = 5      # number of runs before scrollbar appears
    collist = ['p0','blo','bhi','res','dres','chi','fixed']
    selected = 0        # index of selected run 
    
    # ======================================================================= #
    def __init__(self,bfit,parent,group):
        """
            Inputs:
                bfit: top level pointer
                parent      pointer to parent object (frame)
                group: number of data group to fit
        """
        
        # initialize
        self.bfit = bfit
        self.data = self.bfit.data
        self.parent = parent
        self.group = group
        self.parlabels = []
        self.parentry = {}
        
    # ======================================================================= #
    def create(self):
        """Create graphics for this object"""
        
        fitframe = ttk.Frame(self.parent)
        self.parent.add(fitframe,text='Group %d' % self.group)
        
        # get list of runs with the group number
        dl = self.bfit.fetch_files.data_lines
        self.runlist = [dl[k].run for k in dl.keys() 
                if dl[k].group.get() == self.group]
        
        # Display run info label 
        ttk.Label(fitframe,text="Run Numbers").grid(column=0,row=0,padx=5)

        # List box for run viewing
        rlist = StringVar(value=tuple(map(str,self.runlist)))
        self.runbox = Listbox(fitframe,height=min(len(self.runlist),self.n_runs_max),
                         width=10,listvariable=rlist,justify=CENTER,
                        selectmode=BROWSE)
        self.runbox.activate(0)
        self.runbox.bind('<<ListboxSelect>>',self.set_display)
        self.runbox.grid(column=0,row=1,rowspan=10)
        
        sbar = ttk.Scrollbar(fitframe,orient=VERTICAL,command=self.runbox.yview)
        self.runbox.configure(yscrollcommand=sbar.set)
        
        if len(self.runlist) > self.n_runs_max:
            sbar.grid(column=1,row=1,sticky=(N,S),rowspan=10)
        else:
            ttk.Label(fitframe,text=" ").grid(column=1,row=1,padx=5)
        
        # Parameter input labels
        c = 2
        ttk.Label(fitframe,text='Parameter').grid(      column=c,row=0,padx=5); c+=1
        ttk.Label(fitframe,text='Initial Value').grid(  column=c,row=0,padx=5); c+=1
        ttk.Label(fitframe,text='Low Bound').grid(      column=c,row=0,padx=5); c+=1
        ttk.Label(fitframe,text='High Bound').grid(     column=c,row=0,padx=5); c+=1
        ttk.Label(fitframe,text='Result').grid(         column=c,row=0,padx=5); c+=1
        ttk.Label(fitframe,text='Result Error').grid(   column=c,row=0,padx=5); c+=1
        ttk.Label(fitframe,text='ChiSq').grid(          column=c,row=0,padx=5); c+=1
        ttk.Label(fitframe,text='Fixed').grid(          column=c,row=0,padx=5); c+=1
        
        # save
        self.fitframe = fitframe
        
    # ======================================================================= #
    def get_new_parameters(self,runlist):
        """
            Fetch initial parameters from fitter, set to data.    
            
            runlist: list of run numbers to set new parameters for. 
        """
        
        # get pointer to fit files object
        fit_files = self.bfit.fit_files
        fitter = fit_files.fitter
        ncomp = fit_files.n_component.get()
        fn_title = fit_files.fit_function_title.get()
        
        # get list of parameter names
        plist = fitter.gen_param_names(fn_title,ncomp)
        for run in runlist:
            
            # get init values
            values = fitter.gen_init_par(fn_title,ncomp,self.data[run].bd)
            
            # set to data
            self.data[run].set_fitpar(values)
    
        return plist
        
    # ======================================================================= #
    def get_selected_run(self):
        """Get the run number of the selected run"""
        
        try:
            self.selected = self.runbox.curselection()[0]
        except IndexError:
            self.selected = 0 
            
        return self.runlist[self.selected]
        
    # ======================================================================= #
    def populate_param(self):
        """Populate the list of parameters"""
        
        # get list of parameters and initial values
        try:
            plist = self.get_new_parameters(self.runlist)
        except KeyError:
            return
        finally:
            for label in self.parlabels:
                label.destroy()
            for k in self.parentry.keys():
                for p in self.parentry[k]:
                    self.parentry[k][p][1].destroy()
        
        
        # make parameter input fields ---------------------------------------
        
        # labels
        c = 2
        
        self.parlabels = []     # track all labels and inputs
        for i,p in enumerate(plist):
            self.parlabels.append(ttk.Label(self.fitframe,text=p,justify=LEFT))
            self.parlabels[-1].grid(column=c,row=1+i,padx=5,sticky=E)
        
        # get data of selected run
        run = self.get_selected_run()
        fitdat = self.data[run]
        
        # input values: initial parameters
        r = 0
        for p in plist:         # iterate parameter names
            c = 2   # gridding column         
            r += 1  # gridding row         
            self.parentry[p] = {}
            for i in range(3):
                c += 1
                
                value = StringVar()
                entry = ttk.Entry(self.fitframe,textvariable=value,width=10)
                entry.insert(0,str(fitdat.fitpar[self.collist[i]][p]))
                entry.grid(column=c,row=r,padx=5,sticky=E)
                self.parentry[p][self.collist[i]] = (value,entry)
            
            # do results
            c += 1
            par_val = StringVar()
            par = ttk.Entry(self.fitframe,textvariable=par_val,width=15)
            par['state'] = 'readonly'
            par['foreground'] = 'black'
            
            dpar_val = StringVar()
            dpar = ttk.Entry(self.fitframe,textvariable=dpar_val,width=10)
            dpar['state'] = 'readonly'
            dpar['foreground'] = 'black'
                                     
            par. grid(column=c,row=r,padx=5,sticky=E); c += 1
            dpar.grid(column=c,row=r,padx=5,sticky=E); c += 1

            # do chi only once
            if r == 1:
                chi_val = StringVar()
                chi = ttk.Entry(self.fitframe,textvariable=chi_val,width=7)
                chi['state'] = 'readonly'
                chi['foreground'] = 'black'
                
                chi.grid(column=c,row=r,padx=5,sticky=E,rowspan=len(plist)); 
            c += 1
            
            # save ttk.Entry objects in dictionary [parname][colname]
            self.parentry[p][self.collist[3]] = (par_val,par)
            self.parentry[p][self.collist[4]] = (dpar_val,dpar)
            self.parentry[p][self.collist[5]] = (chi_val,chi)
            
            # do fixed box
            value = BooleanVar()
            entry = ttk.Checkbutton(self.fitframe,text='',\
                                     variable=value,onvalue=True,offvalue=False)
            entry.grid(column=c,row=r,padx=5,sticky=E)
            self.parentry[p][self.collist[6]] = (value,entry)
        
        # set parameters
        self.set_display()
        
    # ======================================================================= #
    def set_display(self,*args):
        """Set initial parameters and fit results in display to that of selected run"""
        
        # INITIAL PARAMETERS
        
        # get data that is currently there, possibly for whole group
        run = self.runlist[self.selected]
        if self.bfit.fit_files.set_as_group.get():        
            fitdat_old = [self.data[r] for r in self.runlist]
        else:
            fitdat_old = [self.data[run]]
        
        # get run number of new selected run
        run = self.get_selected_run()
        fitdat_new = self.data[run]
    
    
        for p in self.parentry.keys():  # parentry = [parname][colname][value,entry]
            for i in range(3):          # iterate input columns
                col = self.collist[i]   # column title
        
                # get new initial parameters
                if len(fitdat_new.fitpar[col].keys()) == 0:
                    self.get_new_parameters([run])
                    
                # get data of old entry
                for d in fitdat_old:
                    d.fitpar[col][p] = float(self.parentry[p][col][0].get())
                
                # set values of new data
                self.parentry[p][col][0].set(
                            ("%"+".%df" % self.bfit.rounding) % \
                            fitdat_new.fitpar[col][p])
                
            # get fixed status of old data then set to new
            try:
                for d in fitdat_old:
                    d.fitpar['fixed'][p] = self.parentry[p]['fixed'][0].get()
            except KeyError:
                pass
            
            try:
                self.parentry[p]['fixed'][0].set(fitdat_new.fitpar['fixed'][p])
            except KeyError:
                fitdat_new.fitpar['fixed'][p] = False
                self.parentry[p]['fixed'][0].set(False)
    
        
        # FIT RESULTS    
        
        # Set up variables
        displays = self.parentry
        
        # get run number of selected run
        run = self.get_selected_run()
        
        try:
            data = self.data[run]
        except KeyError:
            return
            
        try:
            chi = data.chi
        except AttributeError:
            return 
        
        # display
        for parname in data.fitpar['res'].keys():
            disp = displays[parname]
            showstr = "%"+".%df" % self.bfit.rounding
            disp['res'][0].set(showstr % data.fitpar['res'][parname])
            disp['dres'][0].set(showstr % data.fitpar['dres'][parname])
            disp['chi'][0].set(showstr % chi)
         
    # ======================================================================= #
    def set_run_color(self):
        """On fit, set the color of the line in the run number select."""

        runlist = map(int,self.runbox.get(0,self.runbox.size()))
        
        for i,r in enumerate(runlist):
            if self.data[r].chi > self.bfit.fit_files.chi_threshold:
                self.runbox.itemconfig(i, {'bg':'red'})
            else:
                self.runbox.itemconfig(i, {'bg':'white'})

    # ======================================================================= #
    def update(self):
        """Update tab with new data"""
        
        # get list of runs with the group number
        dl = self.bfit.fetch_files.data_lines
        self.runlist = [dl[k].run for k in dl.keys() 
                if dl[k].group.get() == self.group]
        
        # List box for run viewing
        rlist = StringVar(value=tuple(map(str,self.runlist)))
        self.runbox.config(height=min(len(self.runlist),self.n_runs_max))
        self.runbox.config(listvariable=rlist)
        self.runbox.activate(0)
        
        
        
        
        



# Cython functions for fast numerical integration 
# Derek Fujimoto
# October 2017
#
# Based on the work of John D. Cook
# https://www.johndcook.com/blog/double_exponential_integration/
# Note: to see slow lines write 'cython integrator.pyx -a --cplus'

# To build write: python setup_integrator.py build_ext --inplace
# assuming the cython seup file is called setup_integrator.

# python3 setup_integrator.py build_ext --inplace

cimport cython
import numpy as np
cimport numpy as np
from libc.math cimport exp 

# ========================================================================== #
# Integration functions import
cdef extern from 'integration_fns.h':
    cdef cppclass Integrator:
        double lifetime;
        Integrator(double);
        double StrExp(double,double,double,double) except +;
        double MixedStrExp(double,double,double,double,double,double,double) except +;
        
# =========================================================================== #
# Integrator class
cdef class PulsedFns:
    cdef double life            # probe lifetime in s
    cdef double pulse_len       # length of beam on in s

# =========================================================================== #
    def __init__(self,lifetime,pulse_len):
        """
            Inputs:
                lifetime: probe lifetime in s
                pulse_len: beam on pulse length in s
        """
        self.life = lifetime
        self.pulse_len = pulse_len

# =========================================================================== #
    @cython.boundscheck(False)  # some speed up in exchange for instability
    cpdef exp(self,np.ndarray[double, ndim=1] time,double Lambda,
                   double amp):
        """
            Pulsed exponential for an array of times. Efficient c-speed looping 
            and indexing. 
            
            Inputs: 
                time: array of times
                Lambda: 1/T1 in s^-1
                Amp: amplitude
                
            Outputs: 
                np.array of values for the puslsed stretched exponential. 
        """
        
        # Variable definitions
        cdef int n = time.shape[0]
        cdef int i
        cdef double t
        cdef np.ndarray[double, ndim=1] out = np.zeros(n)
        cdef double life = self.life
        cdef double pulse_len = self.pulse_len
        cdef double prefac
        cdef double lambda1
        cdef double afterfactor
        
        # precalculations 
        lambda1 = Lambda+1./life
        prefac = amp/(lambda1*life)
        afterfactor = (1-np.exp(-lambda1*pulse_len))/(1-np.exp(-pulse_len/life))
        
        # Calculate pulsed exponential
        for i in range(n):    
            
            # get some useful values: time, normalization
            t = time[i]
        
            # during pulse
            if t<pulse_len:
                out[i] = prefac*(1-np.exp(-lambda1*t))/(1-np.exp(-t/life))
            
            # after pulse
            else:
                out[i] = prefac*afterfactor*np.exp(-Lambda*(t-pulse_len))
        
        return out

# =========================================================================== #
    @cython.boundscheck(False)  # some speed up in exchange for instability
    cpdef str_exp(self,np.ndarray[double, ndim=1] time,double Lambda, 
                         double Beta, double amp):
        """
            Pulsed stretched exponential for an array of times. Efficient 
            c-speed looping and indexing. 
            
            Inputs: 
                time: array of times
                Lambda: 1/T1 in s^-1
                Beta: stretching factor
                Amp: amplitude
                
            Outputs: 
                np.array of values for the puslsed stretched exponential. 
        """
        
        # Variable definitions
        cdef double out
        cdef int n = time.shape[0]
        cdef int i
        cdef double t, x
        cdef np.ndarray[double, ndim=1] out_arr = np.zeros(n)
        cdef double prefac
        cdef double prefac_post
        cdef double life = self.life
        cdef double pulse_len = self.pulse_len
        
        prefac_post = life*(1.-exp(-pulse_len/life))
        
        # Calculate pulsed str. exponential
        for i in range(n):    
            
            # get some useful values: time, normalization
            t = time[i]
                
            # make integrator
            intr = new Integrator(life)
            
            # during pulse
            if t<pulse_len:
                prefac = life*(1.-exp(-t/life))
                x = intr.StrExp(t,t,Lambda,Beta)
                out = amp*x/prefac
            
            # after pulse
            else:
                x = intr.StrExp(t,pulse_len,Lambda,Beta)
                out = amp*x*exp((t-pulse_len)/life)/prefac_post
            
            # save result
            out_arr[i] = out
        
        return out_arr

# =========================================================================== #
    @cython.boundscheck(False)  # some speed up in exchange for instability
    cpdef mixed_str_exp(self,np.ndarray[double, ndim=1] time, 
                double Lambda1, double Beta1, double Lambda2, double Beta2,
                double alpha, double amp):
        """
            Pulsed stretched exponential for an array of times. Efficient 
            c-speed looping and indexing. 
            
            Inputs: 
                time: array of times
                pulse_len: pulse length
                Lambda: 1/T1 in s^-1
                Beta: stretching factor
                alpha: mixing 0 < alpha < 1
                
            Outputs: 
                np.array of values for the puslsed stretched exponential. 
        """
        
        # Variable definitions
        cdef double out
        cdef int n = time.shape[0]
        cdef int i
        cdef double t
        cdef np.ndarray[double, ndim=1] out_arr = np.zeros(n)
        cdef double prefac
        cdef double prefac_post
        cdef double life = self.life
        cdef double pulse_len = self.pulse_len
        
        prefac_post = life*(1.-exp(-pulse_len/life))
        
        # Calculate pulsed str. exponential
        for i in range(n):    
            
            # get some useful values: time, normalization
            t = time[i]
            
            # make integrator
            intr = new Integrator(life)
            
            # during pulse
            if t<pulse_len:
                prefac = life*(1.-exp(-t/life))
                x = intr.MixedStrExp(t,t,Lambda1,Beta1,Lambda2,Beta2,alpha)
                out = amp*x/prefac
            
            # after pulse
            else:
                x = intr.MixedStrExp(t,pulse_len,Lambda1,Beta1,Lambda2,Beta2,
                                     alpha)
                out = amp/prefac_post*x*exp((t-pulse_len)/life)
            
            # save result
            out_arr[i] = out
        
        return out_arr
